@@ -64,6 +64,8 @@ struct SerdReaderImpl {
 	SerdPrefixSink    prefix_sink;
 	SerdStatementSink statement_sink;
 	SerdEndSink       end_sink;
+	SerdErrorSink     error_sink;
+	void*             error_handle;
 	Ref               rdf_first;
 	Ref               rdf_rest;
 	Ref               rdf_nil;
@@ -78,7 +80,9 @@ struct SerdReaderImpl {
 	unsigned          next_id;
 	uint8_t*          read_buf;
 	int32_t           read_head;  ///< Offset into read_buf
+	uint8_t           read_byte;  ///< 1-byte 'buffer' used when not paging
 	bool              from_file;  ///< True iff reading from @ref fd
+	bool              paging;  ///< True iff reading a page at a time
 	bool              eof;
 	bool              seen_genid;
 #ifdef SERD_STACK_CHECK
@@ -88,13 +92,14 @@ struct SerdReaderImpl {
 };
 
 static int
-error(SerdReader* reader, const char* fmt, ...)
+error(SerdReader* reader, SerdStatus st, const char* fmt, ...)
 {
 	va_list args;
 	va_start(args, fmt);
-	fprintf(stderr, "error: %s:%u:%u: ",
-	        reader->cur.filename, reader->cur.line, reader->cur.col);
-	vfprintf(stderr, fmt, args);
+	const SerdError e = {
+		st, reader->cur.filename, reader->cur.line, reader->cur.col, fmt, &args
+	};
+	serd_error(reader->error_sink, reader->error_handle, &e);
 	va_end(args);
 	return 0;
 }
@@ -124,14 +129,19 @@ static inline uint8_t
 eat_byte_safe(SerdReader* reader, const uint8_t byte)
 {
 	assert(peek_byte(reader) == byte);
-	++reader->read_head;
 	switch (byte) {
 	case '\0': reader->eof = true; break;
 	case '\n': ++reader->cur.line; reader->cur.col = 0; break;
 	default:   ++reader->cur.col;
 	}
 
-	if (reader->from_file && (reader->read_head == SERD_PAGE_SIZE)) {
+	if (reader->from_file && !reader->paging) {
+		const int c = fgetc(reader->fd);
+		reader->read_byte = (c == EOF) ? 0 : (uint8_t)c;
+		if (c == EOF) {
+			reader->eof = true;
+		}
+	} else if (++reader->read_head == SERD_PAGE_SIZE && reader->paging) {
 		page(reader);
 	}
 	return byte;
@@ -142,7 +152,8 @@ eat_byte_check(SerdReader* reader, const uint8_t byte)
 {
 	const uint8_t c = peek_byte(reader);
 	if (c != byte) {
-		return error(reader, "expected `%c', not `%c'\n", byte, c);
+		return error(reader, SERD_ERR_BAD_SYNTAX,
+		             "expected `%c', not `%c'\n", byte, c);
 	}
 	return eat_byte_safe(reader, byte);
 }
@@ -264,7 +275,8 @@ read_hex(SerdReader* reader)
 	if (in_range(c, 0x30, 0x39) || in_range(c, 0x41, 0x46)) {
 		return eat_byte_safe(reader, c);
 	} else {
-		return error(reader, "illegal hexadecimal digit `%c'\n", c);
+		return error(reader, SERD_ERR_BAD_SYNTAX,
+		             "invalid hexadecimal digit `%c'\n", c);
 	}
 }
 
@@ -291,7 +303,8 @@ read_hex_escape(SerdReader* reader, unsigned length, Ref dest)
 	} else if (c < 0x00110000) {
 		size = 4;
 	} else {
-		error(reader, "unicode character 0x%X out of range\n", c);
+		error(reader, SERD_ERR_BAD_SYNTAX,
+		      "unicode character 0x%X out of range\n", c);
 		push_replacement(reader, dest);
 		return true;
 	}
@@ -391,20 +404,20 @@ read_ucharacter_escape(SerdReader* reader, Ref dest)
 static inline SerdStatus
 bad_char(SerdReader* reader, Ref dest, const char* fmt, uint8_t c)
 {
-	error(reader, fmt, c);
+	error(reader, SERD_ERR_BAD_SYNTAX, fmt, c);
 	push_replacement(reader, dest);
 
 	// Skip bytes until the next start byte
-	for (uint8_t c = peek_byte(reader); (c & 0x80);) {
-		eat_byte_safe(reader, c);
-		c = peek_byte(reader);
+	for (uint8_t b = peek_byte(reader); (b & 0x80);) {
+		eat_byte_safe(reader, b);
+		b = peek_byte(reader);
 	}
 
 	return SERD_SUCCESS;
 }
 
 static SerdStatus
-read_utf8_character(SerdReader* reader, Ref dest, const uint8_t c)
+read_utf8_character(SerdReader* reader, Ref dest, uint8_t c)
 {
 	unsigned size = 1;
 	if ((c & 0xE0) == 0xC0) {  // Starts with `110'
@@ -447,7 +460,7 @@ read_character(SerdReader* reader, Ref dest)
 	const uint8_t c = peek_byte(reader);
 	assert(c != '\\');  // Only called from methods that handle escapes first
 	if (c == '\0') {
-		error(reader, "unexpected end of file\n", c);
+		error(reader, SERD_ERR_BAD_SYNTAX, "unexpected end of input\n", c);
 		return SERD_ERR_BAD_SYNTAX;
 	} else if (c < 0x20) {
 		return bad_char(reader, dest,
@@ -486,7 +499,8 @@ read_lcharacter(SerdReader* reader, Ref dest, SerdNodeFlags* flags)
 		if (read_scharacter_escape(reader, dest, flags)) {
 			return SERD_SUCCESS;
 		} else {
-			error(reader, "illegal escape `\\%c'\n", peek_byte(reader));
+			error(reader, SERD_ERR_BAD_SYNTAX,
+			      "invalid escape `\\%c'\n", peek_byte(reader));
 			return SERD_ERR_BAD_SYNTAX;
 		}
 	case 0xA: case 0xD:
@@ -510,7 +524,8 @@ read_scharacter(SerdReader* reader, Ref dest, SerdNodeFlags* flags)
 		if (read_scharacter_escape(reader, dest, flags)) {
 			return SERD_SUCCESS;
 		} else {
-			error(reader, "illegal escape `\\%c'\n", peek_byte(reader));
+			error(reader, SERD_ERR_BAD_SYNTAX,
+			      "invalid escape `\\%c'\n", peek_byte(reader));
 			return SERD_ERR_BAD_SYNTAX;
 		}
 	case '\"':
@@ -532,7 +547,8 @@ read_ucharacter(SerdReader* reader, Ref dest)
 		if (read_ucharacter_escape(reader, dest)) {
 			return SERD_SUCCESS;
 		} else {
-			error(reader, "illegal escape `\\%c'\n", peek_byte(reader));
+			error(reader, SERD_ERR_BAD_SYNTAX,
+			      "invalid escape `\\%c'\n", peek_byte(reader));
 			return SERD_FAILURE;
 		}
 	case '>':
@@ -548,7 +564,7 @@ read_comment(SerdReader* reader)
 {
 	eat_byte_safe(reader, '#');
 	uint8_t c;
-	while (((c = peek_byte(reader)) != 0xA) && (c != 0xD)) {
+	while (((c = peek_byte(reader)) != 0xA) && (c != 0xD) && c) {
 		eat_byte_safe(reader, c);
 	}
 }
@@ -669,7 +685,7 @@ static inline uchar
 read_nameStartChar(SerdReader* reader)
 {
 	const uint8_t c = peek_byte(reader);
-	if (c == '_' || is_alpha(c)) {  // TODO: not strictly correct
+	if (c == '_' || is_alpha(c) || is_digit(c)) {  // TODO: Not correct
 		return eat_byte_safe(reader, c);
 	}
 	return 0;
@@ -700,7 +716,7 @@ read_prefixName(SerdReader* reader, Ref dest)
 {
 	uint8_t c = peek_byte(reader);
 	if (c == '_') {
-		error(reader, "unexpected `_'\n");
+		error(reader, SERD_ERR_BAD_SYNTAX, "unexpected `_'\n");
 		return pop_node(reader, dest);
 	}
 	TRY_RET(c = read_nameStartChar(reader));
@@ -734,7 +750,7 @@ read_language(SerdReader* reader)
 {
 	uint8_t c = peek_byte(reader);
 	if (!in_range(c, 'a', 'z')) {
-		return error(reader, "unexpected `%c'\n", c);
+		return error(reader, SERD_ERR_BAD_SYNTAX, "unexpected `%c'\n", c);
 	}
 	Ref ref = push_node(reader, SERD_LITERAL, "", 0);
 	push_byte(reader, ref, eat_byte_safe(reader, c));
@@ -788,7 +804,7 @@ read_0_9(SerdReader* reader, Ref str, bool at_least_one)
 	uint8_t c;
 	if (at_least_one) {
 		if (!is_digit((c = peek_byte(reader)))) {
-			return error(reader, "expected digit\n");
+			return error(reader, SERD_ERR_BAD_SYNTAX, "expected digit\n");
 		}
 		push_byte(reader, str, eat_byte_safe(reader, c));
 	}
@@ -901,7 +917,7 @@ except:
 }
 
 inline static bool
-is_token_end(const uint8_t c)
+is_token_end(uint8_t c)
 {
 	switch (c) {
 	case 0x9: case 0xA: case 0xD: case 0x20: case '\0':
@@ -950,7 +966,8 @@ read_nodeID(SerdReader* reader)
 	                    reader->bprefix ? (char*)reader->bprefix : "",
 	                    reader->bprefix_len);
 	if (!read_name(reader, ref)) {
-		return error(reader, "illegal character at start of name\n");
+		return error(reader, SERD_ERR_BAD_SYNTAX,
+		             "invalid character at start of name\n");
 	}
 	if (reader->syntax == SERD_TURTLE) {
 		const char* const buf = (const char*)deref(reader, ref)->buf;
@@ -958,8 +975,8 @@ read_nodeID(SerdReader* reader)
 			memcpy((char*)buf, "docid", 5);  // Prevent clash
 			reader->seen_genid = true;
 		} else if (reader->seen_genid && !strncmp(buf, "docid", 5)) {
-			error(reader, "found both `genid' and `docid' blank IDs\n");
-			error(reader, "resolve this with a blank ID prefix\n");
+			error(reader, SERD_ERR_ID_CLASH,
+			      "found both `genid' and `docid' IDs, prefix required\n");
 			return pop_node(reader, ref);
 		}
 	}
@@ -1032,7 +1049,7 @@ read_blank(SerdReader* reader, ReadContext ctx, bool subject, Ref* dest)
 	case '(':
 		return read_collection(reader, ctx, dest);
 	default:
-		return error(reader, "illegal blank node\n");
+		return error(reader, SERD_ERR_BAD_SYNTAX, "invalid blank node\n");
 	}
 }
 
@@ -1304,7 +1321,7 @@ read_directive(SerdReader* reader)
 	switch (peek_byte(reader)) {
 	case 'b': return read_base(reader);
 	case 'p': return read_prefixID(reader);
-	default:  return error(reader, "illegal directive\n");
+	default:  return error(reader, SERD_ERR_BAD_SYNTAX, "invalid directive\n");
 	}
 }
 
@@ -1359,6 +1376,8 @@ serd_reader_new(SerdSyntax        syntax,
 	me->prefix_sink      = prefix_sink;
 	me->statement_sink   = statement_sink;
 	me->end_sink         = end_sink;
+	me->error_sink       = NULL;
+	me->error_handle     = NULL;
 	me->default_graph    = SERD_NODE_NULL;
 	me->fd               = 0;
 	me->stack            = serd_stack_new(SERD_PAGE_SIZE);
@@ -1383,6 +1402,16 @@ serd_reader_new(SerdSyntax        syntax,
 	return me;
 }
 
+SERD_API
+void
+serd_reader_set_error_sink(SerdReader*   reader,
+                           SerdErrorSink error_sink,
+                           void*         error_handle)
+{
+	reader->error_sink   = error_sink;
+	reader->error_handle = error_handle;
+}
+
 SERD_API
 void
 serd_reader_free(SerdReader* reader)
@@ -1458,14 +1487,17 @@ static void
 skip_bom(SerdReader* me)
 {
 	const uint8_t* const b = me->read_buf;
-	if (b[0] == 0xEF && b[1] == 0xBB && b[2] == 0xBF) {
+	if (me->paging && b[0] == 0xEF && b[1] == 0xBB && b[2] == 0xBF) {
 		me->read_head += 3;
 	}
 }
 
 SERD_API
 SerdStatus
-serd_reader_read_file_handle(SerdReader* me, FILE* file, const uint8_t* name)
+serd_reader_start_stream(SerdReader*    me,
+                         FILE*          file,
+                         const uint8_t* name,
+                         bool           bulk)
 {
 	const Cursor cur = { name, 1, 1 };
 	me->fd        = file;
@@ -1473,19 +1505,62 @@ serd_reader_read_file_handle(SerdReader* me, FILE* file, const uint8_t* name)
 	me->cur       = cur;
 	me->from_file = true;
 	me->eof       = false;
-	me->read_buf  = (uint8_t*)serd_bufalloc(SERD_PAGE_SIZE);
+	me->paging    = bulk;
 
-	memset(me->read_buf, '\0', SERD_PAGE_SIZE);
-
-	SerdStatus st = page(me);
-	if (!st) {
+	if (bulk) {
+		me->read_buf = (uint8_t*)serd_bufalloc(SERD_PAGE_SIZE);
+		memset(me->read_buf, '\0', SERD_PAGE_SIZE);
+		SerdStatus st = page(me);
+		if (st) {
+			serd_reader_end_stream(me);
+			return st;
+		}
 		skip_bom(me);
-		st = read_turtleDoc(me) ? SERD_SUCCESS : SERD_ERR_UNKNOWN;
+	} else {
+		me->read_buf  = &me->read_byte;
+		me->read_byte = 0;  // Don't read to avoid potentially blocking
 	}
 
-	free(me->read_buf);
+	return SERD_SUCCESS;
+}
+
+SERD_API
+SerdStatus
+serd_reader_read_chunk(SerdReader* me)
+{
+	if (!me->read_byte) {
+		// Read initial byte
+		const int c = fgetc(me->fd);
+		me->read_byte = (c == EOF) ? 0 : (uint8_t)c;
+		if (c == EOF) {
+			me->eof = true;
+			return SERD_FAILURE;
+		}
+	}
+	return read_statement(me) ? SERD_SUCCESS : SERD_FAILURE;
+}
+
+SERD_API
+SerdStatus
+serd_reader_end_stream(SerdReader* me)
+{
+	if (me->paging) {
+		free(me->read_buf);
+	}
 	me->fd       = 0;
 	me->read_buf = NULL;
+	return SERD_SUCCESS;
+}
+
+SERD_API
+SerdStatus
+serd_reader_read_file_handle(SerdReader* me, FILE* file, const uint8_t* name)
+{
+	SerdStatus st = serd_reader_start_stream(me, file, name, true);
+	if (!st) {
+		st = read_turtleDoc(me) ? SERD_SUCCESS : SERD_ERR_UNKNOWN;
+		serd_reader_end_stream(me);
+	}
 	return st;
 }
 
@@ -1499,6 +1574,7 @@ serd_reader_read_string(SerdReader* me, const uint8_t* utf8)
 	me->read_head = 0;
 	me->cur       = cur;
 	me->from_file = false;
+	me->paging    = false;
 	me->eof       = false;
 
 	skip_bom(me);
